# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
import time
from typing import Dict, Optional

from PyQt5.QtCore import QLocale, QMutex
from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.authorities import AuthorityStatus
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import AccountEvent, CurrencyEvent
from tikka.domains.entities.identity import IdentityStatus
from tikka.domains.entities.node import Node
from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.interfaces.adapters.network.node.smiths import NodeSmithsException
from tikka.slots.pyqt.entities.constants import (
    ICON_LOADER,
    SMITH_CERTIFY_SELECTED_IDENTITY_INDEX,
    SMITH_INVITE_SELECTED_ACCOUNT_ADDRESS,
    SMITH_SELECTED_ACCOUNT_ADDRESS,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.widgets.smith_rc import Ui_SmithWidget
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class SmithWidget(QWidget, Ui_SmithWidget):
    """
    SmithWidget class
    """

    DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST = 6

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ) -> None:
        """
        Init SmithWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        self.account: Optional[Account] = None
        self.smith: Optional[Smith] = None
        self.authority_status: AuthorityStatus = AuthorityStatus.OFFLINE
        self.smith_certification_names: Dict[int, str] = {}

        self.invite_account: Optional[Account] = None
        self.certify_smith: Optional[Smith] = None
        self.current_node = self.application.nodes.get(
            self.application.nodes.get_current_url()
        )

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        ##############################
        # ASYNC METHODS
        ##############################
        self.fetch_all_from_network_async_qworker = AsyncQWorker(
            self.fetch_all_from_network, self.mutex
        )
        self.fetch_all_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_all_from_network
        )
        self.fetch_certification_names_from_network_async_qworker = AsyncQWorker(
            self.fetch_certification_names_from_network, self.mutex
        )
        self.fetch_certification_names_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_certification_names_from_network
        )
        self.rotate_keys_async_worker = AsyncQWorker(
            self.network_rotate_keys, self.mutex
        )
        self.rotate_keys_async_worker.finished.connect(
            self._on_network_rotate_keys_finished
        )
        self.publish_keys_async_worker = AsyncQWorker(
            self.network_publish_keys, self.mutex
        )
        self.publish_keys_async_worker.finished.connect(
            self._on_network_publish_keys_finished
        )
        self.invite_member_async_qworker = AsyncQWorker(
            self.network_invite_member, self.mutex
        )
        self.invite_member_async_qworker.finished.connect(
            self._on_network_invite_member_finished
        )
        self.accept_invitation_async_worker = AsyncQWorker(
            self.network_accept_invitation, self.mutex
        )
        self.accept_invitation_async_worker.finished.connect(
            self._on_network_invite_member_finished
        )
        self.certify_smith_async_worker = AsyncQWorker(
            self.network_certify_smith, self.mutex
        )
        self.certify_smith_async_worker.finished.connect(
            self._on_network_certify_smith_finished
        )
        self.go_online_async_worker = AsyncQWorker(self.network_go_online, self.mutex)
        self.go_online_async_worker.finished.connect(
            self._on_network_go_online_finished
        )
        self.go_offline_async_worker = AsyncQWorker(self.network_go_offline, self.mutex)
        self.go_offline_async_worker.finished.connect(
            self._on_network_go_offline_finished
        )

        # events
        self.accountComboBox.activated.connect(self.on_account_combobox_index_changed)
        self.refreshSmithButton.clicked.connect(self._on_refresh_smith_button_clicked)
        self.rotateKeysButton.clicked.connect(self._on_rotate_key_button_clicked)
        self.publishKeysButton.clicked.connect(self._on_publish_key_button_clicked)
        self.goOnlineButton.clicked.connect(self._go_online_button_clicked)
        self.goOfflineButton.clicked.connect(self._go_offline_button_clicked)
        self.acceptInvitationButton.clicked.connect(
            self._on_accept_invitation_button_clicked
        )
        self.inviteButton.clicked.connect(self._on_invite_member_button_clicked)
        self.inviteAccountComboBox.activated.connect(
            self.on_invite_account_combobox_index_changed
        )
        self.certifyButton.clicked.connect(self._on_certify_smith_button_clicked)
        self.certifySmithComboBox.activated.connect(
            self.on_certify_smith_combobox_index_changed
        )

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )

        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_ADD, self.on_add_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_DELETE, self.on_delete_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, self.on_update_account_event
        )

        # populate form
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

        self._update_ui()

    def update_account(self, account: Optional[Account]) -> None:
        """
        Update account and smith and authority status and certification list

        :param account: New account
        :return:
        """
        self.account = account
        self.authority_status = AuthorityStatus.OFFLINE

        if self.account is None:
            self.smith = None
        else:
            self.smith = None
            identity_index = self.application.identities.get_index_by_address(
                self.account.address
            )
            if identity_index is not None:
                # get smith status
                self.smith = self.application.smiths.get(identity_index)

            if self.smith is not None:
                self.authority_status = self.application.authorities.get_status(
                    self.smith.identity_index
                )
                if self.application.connections.indexer.is_connected():
                    self.fetch_certification_names_from_network_async_qworker.start()

        # exclude new operating account from dependant list
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

    def init_account_combo_box(self) -> None:
        """
        Init combobox with validated identity accounts (with wallets)

        :return:
        """
        self.accountComboBox.clear()
        self.accountComboBox.addItem("-", userData=None)

        for account in self.application.accounts.get_list():
            identity = self.application.identities.get_by_address(account.address)
            if (
                identity is not None
                and identity.status == IdentityStatus.MEMBER
                and self.application.wallets.exists(account.address)
            ):
                self.accountComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        self.account = None
        preference_account_address_selected = self.application.preferences.get(
            SMITH_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_account_address_selected is not None:
            preference_account_selected = self.application.accounts.get_by_address(
                preference_account_address_selected
            )
            if preference_account_selected is not None:
                index = self.accountComboBox.findData(
                    preference_account_address_selected
                )
                if index > -1:
                    self.accountComboBox.setCurrentIndex(index)
                    self.update_account(preference_account_selected)
                else:
                    self.application.preferences.set(
                        SMITH_SELECTED_ACCOUNT_ADDRESS, None
                    )

    def on_account_combobox_index_changed(self):
        """
        Triggered when account selection is changed

        :return:
        """
        address = self.accountComboBox.currentData()
        if address is not None:
            self.update_account(self.application.accounts.get_by_address(address))
        else:
            self.update_account(None)

        self._update_ui()

        self.application.repository.preferences.set(
            SMITH_SELECTED_ACCOUNT_ADDRESS,
            address,
        )

    def init_invite_account_combo_box(self) -> None:
        """
        Init combobox with validated identity accounts (not smith) to invite to be smith

        :return:
        """
        self.inviteAccountComboBox.clear()
        self.inviteAccountComboBox.addItem("-", userData=None)

        for account in self.application.accounts.get_list():
            if self.account is not None and self.account.address != account.address:
                identity = self.application.identities.get_by_address(account.address)
                if (
                    identity is not None
                    and identity.status == IdentityStatus.MEMBER
                    and not self.application.smiths.exists(identity.index)
                ):
                    self.inviteAccountComboBox.addItem(
                        account.name if account.name is not None else account.address,
                        userData=account.address,
                    )

        self.invite_account = None
        preference_invite_account_address_selected = self.application.preferences.get(
            SMITH_INVITE_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_invite_account_address_selected is not None:
            preference_invite_account_selected = (
                self.application.accounts.get_by_address(
                    preference_invite_account_address_selected
                )
            )
            if preference_invite_account_selected is not None:
                index = self.inviteAccountComboBox.findData(
                    preference_invite_account_address_selected
                )
                if index > -1:
                    self.inviteAccountComboBox.setCurrentIndex(index)
                    self.invite_account = preference_invite_account_selected
                else:
                    self.application.preferences.set(
                        SMITH_INVITE_SELECTED_ACCOUNT_ADDRESS, None
                    )

    def on_invite_account_combobox_index_changed(self):
        """
        Triggered when invite account selection is changed

        :return:
        """
        address = self.inviteAccountComboBox.currentData()
        if address is not None:
            self.invite_account = self.application.accounts.get_by_address(address)
        else:
            self.invite_account = None

        self.application.repository.preferences.set(
            SMITH_INVITE_SELECTED_ACCOUNT_ADDRESS,
            None if self.invite_account is None else self.invite_account.address,
        )
        self._update_ui()

    def init_certify_smith_combo_box(self) -> None:
        """
        Init combobox with smith and pending smith accounts to certify

        :return:
        """
        self.certifySmithComboBox.clear()
        self.certifySmithComboBox.addItem("-", userData=None)

        smiths = [
            smith
            for smith in self.application.smiths.list()
            if (
                smith.status == SmithStatus.SMITH or smith.status == SmithStatus.PENDING
            )
        ]

        for smith in smiths:
            identity = self.application.identities.get(smith.identity_index)
            if identity is not None:
                account = self.application.accounts.get_by_address(identity.address)
                if (
                    account is not None
                    and self.account is not None
                    and account.address != self.account.address
                ):
                    self.certifySmithComboBox.addItem(
                        account.name if account.name is not None else account.address,
                        userData=smith.identity_index,
                    )

        self.certify_smith = None
        preference_certify_smith_identity_index_selected = (
            self.application.preferences.get(SMITH_CERTIFY_SELECTED_IDENTITY_INDEX)
        )
        if preference_certify_smith_identity_index_selected is not None:
            preference_certify_smith_selected = self.application.smiths.get(
                int(preference_certify_smith_identity_index_selected)
            )
            if preference_certify_smith_selected is not None:
                index = self.certifySmithComboBox.findData(
                    preference_certify_smith_identity_index_selected
                )
                if index > -1:
                    self.certifySmithComboBox.setCurrentIndex(index)
                    self.certify_smith = preference_certify_smith_selected
                else:
                    self.application.preferences.set(
                        SMITH_CERTIFY_SELECTED_IDENTITY_INDEX, None
                    )

    def on_certify_smith_combobox_index_changed(self):
        """
        Triggered when certify smith selection is changed

        :return:
        """
        identity_index = self.certifySmithComboBox.currentData()
        if identity_index is not None:
            self.certify_smith = self.application.smiths.get(identity_index)
        else:
            self.certify_smith = None

        self.application.repository.preferences.set(
            SMITH_CERTIFY_SELECTED_IDENTITY_INDEX,
            None if self.certify_smith is None else self.certify_smith.identity_index,
        )
        self._update_ui()

    def _on_rotate_key_button_clicked(self):
        """
        Triggered when user click on rotate keys button

        :return:
        """
        self.rotate_keys_async_worker.start()

    def network_rotate_keys(self):
        """
        Triggered when user click on rotate keys button

        :return:
        """
        self.errorLabel.setText("")
        self.rotateKeysButton.setDisabled(True)
        try:
            self.application.authorities.network_rotate_keys(self.current_node)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

    def _on_network_rotate_keys_finished(self):
        """
        Triggered when network_rotate_keys is finished

        :return:
        """
        self.sessionKeysTextBrowser.setText(self.current_node.session_keys)
        self.errorLabel.setText("")

        self.rotateKeysButton.setEnabled(True)

    def _on_invite_member_button_clicked(self):
        """
        Triggered when user click on invite button

        :return:
        """
        if self.invite_account is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.inviteButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                self.loaderIconLabel.hide()
                return

        self.invite_member_async_qworker.start()

    def network_invite_member(self):
        """
        Send invite smith command to network

        :return:
        """
        identity_index = self.application.identities.get_index_by_address(
            self.invite_account.address
        )
        try:
            self.application.smiths.network_invite_member(
                self.application.wallets.get_keypair(self.account.address),
                identity_index,
            )
        except Exception as exception:
            logging.exception(exception)
            self.errorLabel.setText(self._(str(exception)))
        else:
            try:
                self.application.smiths.network_update_smith(identity_index)
            except Exception as exception:
                logging.exception(exception)
                self.errorLabel.setText(self._(str(exception)))

    def _on_network_invite_member_finished(self):
        """
        Triggered when network_invite_member is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.inviteButton.setEnabled(True)
        self.init_invite_account_combo_box()
        self._update_ui()
        time.sleep(self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST)
        self.fetch_all_from_network_async_qworker.start()

    def _on_accept_invitation_button_clicked(self):
        """
        Triggered when user click on accept invitation button

        :return:
        """
        if self.account is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.acceptInvitationButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                self.loaderIconLabel.hide()
                return
        self.accept_invitation_async_worker.start()

    def network_accept_invitation(self):
        """
        Send accept smith invitation to network

        :return:
        """
        try:
            self.application.smiths.network_accept_invitation(
                self.application.wallets.get_keypair(self.account.address)
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            identity_index = self.application.identities.get_index_by_address(
                self.account.address
            )
            try:
                self.smith = self.application.smiths.network_update_smith(
                    identity_index
                )
            except Exception as exception:
                logging.exception(exception)
                self.errorLabel.setText(self._(str(exception)))

    def _on_network_accept_invitation_finished(self):
        """
        Triggered when network_invite_member is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.acceptInvitationButton.setEnabled(True)
        time.sleep(self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST)
        self.fetch_all_from_network_async_qworker.start()

    def _on_certify_smith_button_clicked(self):
        """
        Triggered when user click on certify button

        :return:
        """
        if self.certify_smith is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.certifyButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                self.loaderIconLabel.hide()
                self.certifyButton.setEnabled(True)
                return

        self.certify_smith_async_worker.start()

    def network_certify_smith(self):
        """
        Send certify smith to network

        :return:
        """
        try:
            self.application.smiths.network_certify(
                self.application.wallets.get_keypair(self.account.address),
                self.certify_smith.identity_index,
            )
        except NodeSmithsException as exception:
            self.errorLabel.setText(self._(str(exception)))
        else:
            try:
                self.application.smiths.network_update_smith(
                    self.certify_smith.identity_index
                )
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                try:
                    self.smith = self.application.smiths.network_update_smith(
                        self.smith.identity_index
                    )
                except Exception as exception:
                    self.errorLabel.setText(self._(str(exception)))
                    logging.exception(exception)

    def _on_network_certify_smith_finished(self):
        """
        Triggered when network_certify_smith is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.certifyButton.setEnabled(True)
        self._update_ui()

        time.sleep(self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST)
        self.fetch_all_from_network_async_qworker.start()

    def _on_publish_key_button_clicked(self):
        """
        Triggered when user click on publish keys button

        :return:
        """

        if self.account is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.publishKeysButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.publish_keys_async_worker.start()

    def network_publish_keys(self):
        """
        Publish keys of local node to network

        :return:
        """
        try:
            self.application.authorities.network_publish_session_keys(
                self.application.wallets.get_keypair(self.account.address),
                self.current_node.session_keys,
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_network_publish_keys_finished(self):
        """
        Triggered when network_publish_keys is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.publishKeysButton.setEnabled(True)

    def _go_online_button_clicked(self):
        """
        Triggered when user click on go online button

        :return:
        """

        if self.account is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.goOnlineButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.go_online_async_worker.start()

    def network_go_online(self):
        """
        Triggered when user click on go online button

        :return:
        """
        try:
            self.application.authorities.network_go_online(
                self.application.wallets.get_keypair(self.account.address)
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

    def _on_network_go_online_finished(self):
        """
        Triggered when network_go_online is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.goOnlineButton.setEnabled(True)

        time.sleep(self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST)
        self.fetch_all_from_network_async_qworker.start()

    def _go_offline_button_clicked(self):
        """
        Triggered when user click on go offline button

        :return:
        """

        if self.account is None:
            return

        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.goOfflineButton.setDisabled(True)

        # if account is locked...
        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.go_offline_async_worker.start()

    def network_go_offline(self):
        """
        Triggered when user click on go offline button

        :return:
        """
        try:
            self.application.authorities.network_go_offline(
                self.application.wallets.get_keypair(self.account.address)
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

    def _on_network_go_offline_finished(self):
        """
        Triggered when network_go_offline is finished

        :return:
        """
        self.loaderIconLabel.hide()
        self.goOfflineButton.setEnabled(True)

        time.sleep(self.DELAY_BEFORE_UPDATE_MEMBERSHIP_STATUS_AFTER_REQUEST)
        self.fetch_all_from_network_async_qworker.start()

    def _on_refresh_smith_button_clicked(self, _):
        """ """
        self.fetch_all_from_network_async_qworker.start()

    def fetch_all_from_network(self):
        """
        Update identities, smiths and authorities from current url connection

        :return:
        """
        self.loader_movie.start()
        self.loaderIconLabel.show()

        addresses = [
            account.address for account in self.application.accounts.get_list()
        ]
        try:
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            identity_indice = self.application.identities.list_indice()
            try:
                self.application.smiths.network_update_smiths(identity_indice)
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                try:
                    self.application.authorities.network_get_all()
                except Exception as exception:
                    self.errorLabel.setText(self._(str(exception)))
                    logging.exception(exception)

    def _on_finished_fetch_all_from_network(self):
        """
        Triggered when AsyncWorker is finished

        :return:
        """
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

        self.refreshSmithButton.setEnabled(True)
        self.loaderIconLabel.hide()

        self._update_ui()

    def fetch_certification_names_from_network(self):
        """
        Load certification identity names from current url connection

        :return:
        """
        if self.smith is not None:
            indice = list(
                set(
                    self.smith.certifications_issued
                    + self.smith.certifications_received
                )
            )
            self.smith_certification_names = (
                self.application.identities.indexer_identities.get_identity_names(
                    indice
                )
            )

    def _on_finished_fetch_certification_names_from_network(self):
        """
        Triggered when AsyncWorker is finished

        :return:
        """
        logging.debug("Smith widget update")

        self._update_ui()

    def _update_ui(self):
        """
        Update node infos in UI

        :return:
        """
        display_smith_status = {
            SmithStatus.INVITED.value: self._("Invited"),
            SmithStatus.PENDING.value: self._("Pending"),
            SmithStatus.SMITH.value: self._("Smith"),
            SmithStatus.EXCLUDED.value: self._("Excluded"),
        }

        # validator node
        if self.current_node.unsafe_api_exposed is True:
            self.validatorNodeGroupBox.show()
        else:
            self.validatorNodeGroupBox.hide()

        self.urlValueLabel.setText(self.current_node.url)

        if self.current_node.session_keys is not None:
            self.sessionKeysTextBrowser.setText(self.current_node.session_keys)
        else:
            self.sessionKeysTextBrowser.setText("")

        # rotate keys only available on localhost via an ssh tunnel or other method...
        self.rotateKeysButton.setEnabled(
            "localhost" in self.application.nodes.get_current_url()
        )

        # disable all buttons
        self.publishKeysButton.setDisabled(True)
        self.inviteButton.setDisabled(True)
        self.inviteAccountComboBox.setDisabled(True)
        self.acceptInvitationButton.setDisabled(True)
        self.certifySmithComboBox.setDisabled(True)
        self.goOnlineButton.setDisabled(True)
        self.goOfflineButton.setDisabled(True)

        if self.account is not None:
            self.refreshSmithButton.setEnabled(True)
            if self.current_node.session_keys is not None:
                self.publishKeysButton.setEnabled(True)
            if self.smith is None:
                pass
            elif self.smith.status == SmithStatus.INVITED:
                self.acceptInvitationButton.setEnabled(True)
            elif self.smith.status == SmithStatus.SMITH:
                self.inviteButton.setEnabled(True)
                if self.inviteAccountComboBox.count() > 1:
                    self.inviteAccountComboBox.setEnabled(True)
                if self.certifySmithComboBox.count() > -1:
                    self.certifySmithComboBox.setEnabled(True)
                self.goOnlineButton.setEnabled(
                    self.authority_status == AuthorityStatus.OFFLINE
                )
                self.goOfflineButton.setEnabled(
                    self.authority_status == AuthorityStatus.ONLINE
                )

        if self.smith is not None:
            status_string = display_smith_status[self.smith.status.value]
            if self.smith.expire_on is not None:
                expire_on_localized_datetime_string = self.locale().toString(
                    self.smith.expire_on,
                    QLocale.dateTimeFormat(self.locale(), QLocale.ShortFormat),
                )
                status_string = (
                    f"{status_string} ({expire_on_localized_datetime_string})"
                )
            self.membershipValueLabel.setText(status_string)

            self.certifiersListWidget.clear()
            for certifier_identity_index in self.smith.certifications_received:
                identity = self.application.identities.get(certifier_identity_index)
                if identity is not None:
                    account = self.application.accounts.get_by_address(identity.address)
                    if account is not None and account.name is not None:
                        self.certifiersListWidget.addItem(account.name)
                else:
                    if certifier_identity_index in self.smith_certification_names:
                        self.certifiersListWidget.addItem(
                            f"{self.smith_certification_names[certifier_identity_index]}#{certifier_identity_index}"
                        )
                    else:
                        self.certifiersListWidget.addItem(
                            f"#{certifier_identity_index}"
                        )

            self.certifiedListWidget.clear()
            for certified_identity_index in self.smith.certifications_issued:
                identity = self.application.identities.get(certified_identity_index)
                if identity is not None:
                    account = self.application.accounts.get_by_address(identity.address)
                    if account is not None and account.name is not None:
                        self.certifiedListWidget.addItem(account.name)
                else:
                    if certified_identity_index in self.smith_certification_names:
                        self.certifiedListWidget.addItem(
                            f"{self.smith_certification_names[certified_identity_index]}#{certified_identity_index}"
                        )
                    else:
                        self.certifiedListWidget.addItem(f"#{certified_identity_index}")
        else:
            self.membershipValueLabel.setText(self._("No"))
            # clear certification list
            self.certifiersListWidget.clear()
            self.certifiedListWidget.clear()

        if self.authority_status == AuthorityStatus.OFFLINE:
            self.authorityValueLabel.setText(self._("No"))
        elif self.authority_status == AuthorityStatus.INCOMING:
            self.authorityValueLabel.setText(self._("Incoming..."))
        elif self.authority_status == AuthorityStatus.ONLINE:
            self.authorityValueLabel.setText(self._("Online"))
        elif self.authority_status == AuthorityStatus.OUTGOING:
            self.authorityValueLabel.setText(self._("Outgoing..."))

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()
        if self.account is not None:
            self.fetch_all_from_network_async_qworker.start()

        self._update_ui()

    def on_connections_event(self, _):
        """
        Triggered when the network connection if connected/disconnected

        :param _: ConnectionsEvent instance
        :return:
        """

        self.current_node = self.application.nodes.get(
            self.application.nodes.get_current_url()
        )
        self.init_account_combo_box()
        self._update_ui()

    def on_add_account_event(self, _):
        """
        Add account is selectors when account is created

        :return:
        """
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

    def on_delete_account_event(self, _):
        """
        Remove account from selector when account is deleted

        :return:
        """
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

        self._update_ui()

    def on_update_account_event(self, _):
        """
        Update account on page and selector when account is updated

        :return:
        """
        self.init_account_combo_box()
        self.init_invite_account_combo_box()
        self.init_certify_smith_combo_box()

        self._update_ui()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    validator_node = Node(url="ws://localhost:9944")
    if application_.nodes.get(validator_node.url) is None:
        application_.nodes.add(validator_node)
    application_.nodes.set_current_url(validator_node.url)
    application_.connections.node.connect(validator_node)
    main_window.setCentralWidget(SmithWidget(application_, QMutex(), main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from tikka.adapters.network.node.connection import NodeConnection
from tikka.domains.entities.events import ConnectionsEvent
from tikka.domains.events import EventDispatcher
from tikka.interfaces.adapters.network.connection import ConnectionInterface
from tikka.interfaces.adapters.network.network import NetworkInterface
from tikka.interfaces.domains.connections import ConnectionsInterface
from tikka.interfaces.entities.Server import ServerInterface


class ConnectionWithEventDispatcher(ConnectionInterface):
    def __init__(
        self, connection: ConnectionInterface, event_dispatcher: EventDispatcher
    ) -> None:
        """
        Init ConnectionWithEventDispatcher instance

        :param connection: EventDispatcher instance
        :param event_dispatcher: EventDispatcher instance
        :return:
        """
        self.connection = connection
        self.event_dispatcher = event_dispatcher

    def connect(self, server: ServerInterface) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.connect.__doc__
        )
        self.connection.connect(server)
        if self.connection.is_connected():
            if isinstance(self.connection, NodeConnection):
                self.event_dispatcher.dispatch_event(
                    ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_NODE_CONNECTED)
                )
            else:
                self.event_dispatcher.dispatch_event(
                    ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED)
                )

    def disconnect(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.disconnect.__doc__
        )
        self.connection.disconnect()
        if not self.connection.is_connected():
            if isinstance(self.connection, NodeConnection):
                self.event_dispatcher.dispatch_event(
                    ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_NODE_DISCONNECTED)
                )
            else:
                self.event_dispatcher.dispatch_event(
                    ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_INDEXER_DISCONNECTED)
                )

    def is_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.is_connected.__doc__
        )
        return self.connection.is_connected()


class Connections(ConnectionsInterface):
    """
    Connections domain class
    """

    def __init__(
        self, network: NetworkInterface, event_dispatcher: EventDispatcher
    ) -> None:
        """
        Init Connections domain instance

        :param network: NetworkInterface instance
        :param event_dispatcher: EventDispatcher instance
        :return:
        """
        super().__init__(network, event_dispatcher)

        self._node = self.network.node.connection
        self._indexer = self.network.indexer.connection

    @property
    def node(self) -> ConnectionInterface:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionsInterface.node.__doc__
        )
        return self._node

    @property
    def indexer(self) -> ConnectionInterface:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionsInterface.indexer.__doc__
        )
        return self._indexer

    def disconnect_all(self):
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionsInterface.disconnect_all.__doc__
        )
        self.node.disconnect()
        self.indexer.disconnect()

    def are_all_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionsInterface.are_all_connected.__doc__
        )
        return self.node.is_connected() and self.indexer.is_connected()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from gql import gql

from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.identities import (
    IndexerIdentitiesException,
    IndexerIdentitiesInterface,
)


class IndexerIdentities(IndexerIdentitiesInterface):
    """
    IndexerIdentities class
    """

    def get_identity_name(self, identity_index: int) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerIdentitiesInterface.get_identity_name.__doc__
        )
        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerIdentitiesException(NetworkConnectionError())

        query = f"""
                query {{
                    identities(
                        filter: {{
                            index: {{
                                equalTo: {identity_index}
                             }}
                        }}
                        ) {{
                            nodes {{
                                name
                            }}
                    }}
                }}
                """

        try:
            result = self.indexer.connection.client.execute(gql(query))
        except Exception as exception:
            raise IndexerIdentitiesException(exception)

        #
        #    {
        #         "identity": [
        #             {
        #                 "name": "vit"
        #             }
        #         ]
        #     }
        #
        return result["identities"]["nodes"][0]["name"]

    def get_identity_names(self, index_list: List[int]) -> Dict[int, str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerIdentitiesInterface.get_identity_names.__doc__
        )
        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerIdentitiesException(NetworkConnectionError())

        query = """
            query GetNamesByIndex($indices: [Int!]!) {
                    identities(filter: { index: { in: $indices } }) {
                        nodes{
                          index
                          name
                        }
                        
                    }
                }
                """

        variables = {"indices": index_list}
        try:
            result = self.indexer.connection.client.execute(
                gql(query), variable_values=variables
            )
        except Exception as exception:
            raise IndexerIdentitiesException(exception)

        #
        #    {
        #   "data": {
        #     "identities": {
        #       "nodes": [
        #         {
        #           "index": 1,
        #           "name": "Alfybe"
        #         },
        #         {
        #           "index": 2,
        #           "name": "AnneAmbles"
        #         },
        #         {
        #           "index": 3,
        #           "name": "Ariane"
        #         }
        #       ]
        #     }
        #   }
        # }
        #
        names = {}
        for row in result["identities"]["nodes"]:
            names[row["index"]] = row["name"]
        return names



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/node_add.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_nodeAddDialog(object):
    def setupUi(self, nodeAddDialog):
        nodeAddDialog.setObjectName("nodeAddDialog")
        nodeAddDialog.resize(731, 241)
        self.verticalLayout = QtWidgets.QVBoxLayout(nodeAddDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.softwareValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.softwareValueLabel.setText("")
        self.softwareValueLabel.setObjectName("softwareValueLabel")
        self.gridLayout.addWidget(self.softwareValueLabel, 1, 1, 1, 1)
        self.versionValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.versionValueLabel.setText("")
        self.versionValueLabel.setObjectName("versionValueLabel")
        self.gridLayout.addWidget(self.versionValueLabel, 2, 1, 1, 1)
        self.urlLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.urlLabel.setFont(font)
        self.urlLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.urlLabel.setObjectName("urlLabel")
        self.gridLayout.addWidget(self.urlLabel, 0, 0, 1, 1)
        self.blockLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.blockLabel.setFont(font)
        self.blockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.blockLabel.setObjectName("blockLabel")
        self.gridLayout.addWidget(self.blockLabel, 4, 0, 1, 1)
        self.versionLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.versionLabel.setFont(font)
        self.versionLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.versionLabel.setObjectName("versionLabel")
        self.gridLayout.addWidget(self.versionLabel, 2, 0, 1, 1)
        self.softwareLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.softwareLabel.setFont(font)
        self.softwareLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.softwareLabel.setObjectName("softwareLabel")
        self.gridLayout.addWidget(self.softwareLabel, 1, 0, 1, 1)
        self.currencyHashLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.currencyHashLabel.setFont(font)
        self.currencyHashLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.currencyHashLabel.setObjectName("currencyHashLabel")
        self.gridLayout.addWidget(self.currencyHashLabel, 5, 0, 1, 1)
        self.currencyHashValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.currencyHashValueLabel.setText("")
        self.currencyHashValueLabel.setObjectName("currencyHashValueLabel")
        self.gridLayout.addWidget(self.currencyHashValueLabel, 5, 1, 1, 1)
        self.peerIDValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.peerIDValueLabel.setText("")
        self.peerIDValueLabel.setObjectName("peerIDValueLabel")
        self.gridLayout.addWidget(self.peerIDValueLabel, 3, 1, 1, 1)
        self.peerIDLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.peerIDLabel.setFont(font)
        self.peerIDLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.peerIDLabel.setObjectName("peerIDLabel")
        self.gridLayout.addWidget(self.peerIDLabel, 3, 0, 1, 1)
        self.urlLineEdit = QtWidgets.QLineEdit(nodeAddDialog)
        self.urlLineEdit.setObjectName("urlLineEdit")
        self.gridLayout.addWidget(self.urlLineEdit, 0, 1, 1, 1)
        self.testButton = QtWidgets.QPushButton(nodeAddDialog)
        self.testButton.setObjectName("testButton")
        self.gridLayout.addWidget(self.testButton, 0, 2, 1, 1)
        self.blockValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.blockValueLabel.setText("")
        self.blockValueLabel.setObjectName("blockValueLabel")
        self.gridLayout.addWidget(self.blockValueLabel, 4, 1, 1, 1)
        self.currencyPrefixLabel = QtWidgets.QLabel(nodeAddDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.currencyPrefixLabel.setFont(font)
        self.currencyPrefixLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.currencyPrefixLabel.setObjectName("currencyPrefixLabel")
        self.gridLayout.addWidget(self.currencyPrefixLabel, 6, 0, 1, 1)
        self.currencyPrefixValueLabel = QtWidgets.QLabel(nodeAddDialog)
        self.currencyPrefixValueLabel.setText("")
        self.currencyPrefixValueLabel.setObjectName("currencyPrefixValueLabel")
        self.gridLayout.addWidget(self.currencyPrefixValueLabel, 6, 1, 1, 1)
        self.loaderIconLabel = QtWidgets.QLabel(nodeAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loaderIconLabel.sizePolicy().hasHeightForWidth())
        self.loaderIconLabel.setSizePolicy(sizePolicy)
        self.loaderIconLabel.setMinimumSize(QtCore.QSize(0, 0))
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setText("")
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.gridLayout.addWidget(self.loaderIconLabel, 1, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(nodeAddDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(nodeAddDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(nodeAddDialog)
        self.buttonBox.accepted.connect(nodeAddDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(nodeAddDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(nodeAddDialog)
        nodeAddDialog.setTabOrder(self.urlLineEdit, self.testButton)

    def retranslateUi(self, nodeAddDialog):
        #  _translate = QtCore.QCoreApplication.translate
        nodeAddDialog.setWindowTitle(_("Add server"))
        self.urlLabel.setText(_("URL"))
        self.blockLabel.setText(_("Block"))
        self.versionLabel.setText(_("Version"))
        self.softwareLabel.setText(_("Software"))
        self.currencyHashLabel.setText(_("Currency Hash"))
        self.peerIDLabel.setText(_("Peer ID"))
        self.testButton.setText(_("Test"))
        self.currencyPrefixLabel.setText(_("Currency Prefix"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from typing import Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtWidgets import QApplication, QMenu, QMessageBox, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.windows.account_derivation_create import (
    AccountDerivationCreateWindow,
)
from tikka.slots.pyqt.windows.account_rename import AccountRenameWindow
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow
from tikka.slots.pyqt.windows.transfer import TransferWindow
from tikka.slots.pyqt.windows.v1_wallet_restore import V1WalletRestoreWindow
from tikka.slots.pyqt.windows.v1_wallet_restore_from_file import (
    V1WalletRestoreFromFileWindow,
)
from tikka.slots.pyqt.windows.wallet_restore import WalletRestoreWindow


class AccountPopupMenu(QMenu):
    """
    AccountPopupMenu class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ):
        """
        Init AccountPopupMenu instance

        :param application: Application instance
        :param account: Account instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)

        self.application = application
        self.account = account
        self.mutex = mutex
        self._ = self.application.translator.gettext

        transfer_to_action = self.addAction(self._("Transfer to"))
        transfer_to_action.triggered.connect(self.transfer_to)
        if self.application.wallets.exists(account.address):
            transfer_from_action = self.addAction(self._("Transfer from"))
            transfer_from_action.triggered.connect(self.transfer_from)

        # menu actions
        copy_address_to_clipboard_action = self.addAction(
            self._("Copy address to clipboard")
        )
        copy_address_to_clipboard_action.triggered.connect(
            self.copy_address_to_clipboard
        )
        if self.application.wallets.exists(account.address):
            if self.application.wallets.is_unlocked(account.address) is False:
                unlock_account_action = self.addAction(self._("Unlock account access"))
                unlock_account_action.triggered.connect(self.unlock_account)
            else:
                lock_account_action = self.addAction(self._("Lock account access"))
                lock_account_action.triggered.connect(self.lock_account)

            wallet_forget_action = self.addAction(self._("Forget the wallet"))
            wallet_forget_action.triggered.connect(self.wallet_forget)
        else:
            wallet_restore_action = self.addAction(self._("Store the wallet"))
            wallet_restore_action.triggered.connect(self.wallet_restore)

        # if account is a root account...
        if self.account.root is None:
            wallet_password_forgotten_action = self.addAction(
                self._("Password forgotten / change")
            )
            wallet_password_forgotten_action.triggered.connect(
                self.wallet_password_forgotten
            )
            create_derived_account_action = self.addAction(
                self._("Create derived account")
            )
            create_derived_account_action.triggered.connect(self.create_derived_account)

        rename_account_action = self.addAction(self._("Rename account"))
        rename_account_action.triggered.connect(self.rename_account)

        forget_account_action = self.addAction(self._("Forget account"))
        forget_account_action.triggered.connect(self.confirm_forget_account)

    def copy_address_to_clipboard(self):
        """
        Copy address of selected row to clipboard

        :return:
        """
        clipboard = QApplication.clipboard()
        clipboard.setText(self.account.address)

    def unlock_account(self):
        """
        Open account unlock window

        :return:
        """
        AccountUnlockWindow(self.application, self.account, self).exec_()

    def transfer_to(self):
        """
        Open transfer window with account as recipient

        :return:
        """
        TransferWindow(
            self.application, self.mutex, None, self.account, parent=self
        ).exec_()

    def transfer_from(self):
        """
        Open transfer window with account as sender

        :return:
        """
        TransferWindow(
            self.application, self.mutex, self.account, None, parent=self
        ).exec_()

    def lock_account(self):
        """
        Lock account

        :return:
        """
        self.application.accounts.lock(self.account)

    def wallet_password_forgotten(self):
        """
        Open wallet restore window to reset wallet password

        :return:
        """
        # if wallet type is V2...
        if self.account.legacy_v1 is False:
            WalletRestoreWindow(self.application, self.account, True).exec_()
        # wallet V1
        else:
            if self.account.file_import is True:
                V1WalletRestoreFromFileWindow(
                    self.application, self.account, True
                ).exec_()
            else:
                V1WalletRestoreWindow(self.application, self.account, True).exec_()

    def wallet_restore(self):
        """
        Open wallet restore window

        :return:
        """
        # if wallet type is V2...
        if self.account.legacy_v1 is False:
            WalletRestoreWindow(self.application, self.account).exec_()
        # wallet V1
        else:
            if self.account.file_import is True:
                V1WalletRestoreFromFileWindow(self.application, self.account).exec_()
            else:
                V1WalletRestoreWindow(self.application, self.account).exec_()

    def wallet_forget(self):
        """
        Forget wallet for this account

        :return:
        """
        # display confirm dialog and get response
        button = QMessageBox.question(
            self,
            self._("Forget wallet"),
            self._("Forget wallet for account {address}?").format(
                address=self.account.address
            ),
        )
        if button == QMessageBox.Yes:
            self.application.accounts.forget_wallet(self.account)

    def create_derived_account(self):
        """
        Open create account derivation window

        :return:
        """
        AccountDerivationCreateWindow(self.application, self.account).exec_()

    def rename_account(self):
        """
        Open rename account window

        :return:
        """
        AccountRenameWindow(self.application, self.account).exec_()

    def confirm_forget_account(self):
        """
        Display confirm dialog then forget account if confirmed

        :return:
        """
        # display confirm dialog and get response
        button = QMessageBox.question(
            self,
            self._("Forget account"),
            self._("Forget account {address} (and derived accounts if any)?").format(
                address=self.account.address
            ),
        )
        if button == QMessageBox.Yes:
            for derived_account in self.application.accounts.get_derivation_accounts(
                self.account.address
            ):
                self.application.accounts.delete(derived_account)
            self.application.accounts.delete(self.account)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = Account("732SSfuwjB7jkt9th1zerGhphs6nknaCBCTozxUcPWPU")

    menu = AccountPopupMenu(application_, account_, QMutex())
    menu.exec_()

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QMutex
from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import QApplication, QListWidgetItem, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import CurrencyEvent
from tikka.domains.entities.node import Node
from tikka.slots.pyqt.entities.constants import (
    ICON_LOADER,
    TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.widgets.technical_committee_rc import (
    Ui_technicalCommitteeWidget,
)
from tikka.slots.pyqt.widgets.technical_committee_proposal import (
    TechnicalCommitteeProposalWidget,
)


class TechnicalCommitteeWidget(QWidget, Ui_technicalCommitteeWidget):
    """
    TechnicalCommitteeWidget class
    """

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ) -> None:
        """
        Init TechnicalCommitteeWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex
        self.proposals_list_current_index = 0

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        ##############################
        # ASYNC METHODS
        ##############################
        self.fetch_members_from_network_async_worker = AsyncQWorker(
            self.fetch_members_from_network, self.mutex
        )
        self.fetch_members_from_network_async_worker.finished.connect(
            self._on_fetch_members_from_network_async_worker_finished
        )
        self.fetch_proposals_from_network_async_worker = AsyncQWorker(
            self.fetch_proposals_from_network, self.mutex
        )
        self.fetch_proposals_from_network_async_worker.finished.connect(
            self._on_fetch_proposals_from_network_async_worker_finished
        )

        # events
        self.accountComboBox.activated.connect(self.on_account_combobox_index_changed)
        self.refreshButton.clicked.connect(self._on_refresh_button_clicked)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )

        self.init_member_list()
        self.init_proposal_list()

    def init_member_list(self) -> None:
        """
        Init list of members

        :return:
        """
        self.membersListWidget.clear()

        for member in self.application.technical_committee.list_members():
            if member.identity_index is not None:
                label = (
                    f"{member.address} - {member.identity_name}#{member.identity_index}"
                )
            else:
                label = f"{member.address}"
            account = self.application.accounts.get_by_address(member.address)
            if account is not None and account.name is not None:
                label += f" - {account.name}"

            self.membersListWidget.addItem(label)

    def init_proposal_list(self) -> None:
        """
        Init list of proposals

        :return:
        """

        scrollbar = self.proposalsListWidget.verticalScrollBar()
        if scrollbar.maximum() > 0:
            scroll_ratio = scrollbar.value() / scrollbar.maximum()
        else:
            scroll_ratio = 0
        current_row = self.proposalsListWidget.currentRow()

        self.proposalsListWidget.clear()

        for proposal in sorted(
            self.application.technical_committee.list_proposals(),
            key=lambda proposal: proposal.voting.index,
        ):
            item = QListWidgetItem(self.proposalsListWidget)
            self.proposalsListWidget.addItem(item)
            widget = TechnicalCommitteeProposalWidget(
                self.application, proposal, self.mutex, parent=self
            )
            item.setSizeHint(widget.minimumSizeHint())

            # connect slot self.update_proposals to proposal item event
            widget.vote_for_proposal_start.connect(self.on_vote_for_proposal_start)
            widget.vote_for_proposal_done.connect(self.update_proposals)

            # Associate the custom widget to the list entry
            self.proposalsListWidget.setItemWidget(item, widget)

        if scrollbar.maximum() > 0:
            new_value = int(scroll_ratio * scrollbar.maximum())
            scrollbar.setValue(new_value)

        if 0 <= current_row < self.proposalsListWidget.count():
            self.proposalsListWidget.setCurrentRow(current_row)

    def init_account_combo_box(self) -> None:
        """
        Init combobox with technical committee member accounts (with wallets)

        :return:
        """
        self.accountComboBox.clear()

        accounts = self.application.accounts.get_list()
        for account in accounts:
            if account.address in [
                member.address
                for member in self.application.technical_committee.list_members()
            ] and self.application.wallets.exists(account.address):
                self.accountComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        preference_account_address_selected = self.application.preferences.get(
            TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_account_address_selected is not None:
            preference_account_selected = self.application.accounts.get_by_address(
                preference_account_address_selected
            )
            if preference_account_selected is not None:
                index = self.accountComboBox.findData(
                    preference_account_address_selected
                )
                if index > -1:
                    self.accountComboBox.setCurrentIndex(index)
                else:
                    self.application.preferences.set(
                        TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS, None
                    )

    def _on_refresh_button_clicked(self, _):
        """
        Refresh data from network when refresh button is clicked

        :param _:
        :return:
        """
        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.fetch_members_from_network_async_worker.start()
        self.fetch_proposals_from_network_async_worker.start()

    def on_account_combobox_index_changed(self):
        """
        Triggered when account selection is changed

        :return:
        """
        address = self.accountComboBox.currentData()

        self.application.repository.preferences.set(
            TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS,
            address,
        )

    def fetch_members_from_network(self):
        """
        Fetch members of technical committee from network

        :return:
        """
        try:
            self.application.technical_committee.network_update_members()
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_fetch_members_from_network_async_worker_finished(self):
        """
        Triggered when async worker is finished

        :return:
        """
        logging.debug("Technical committee widget members update")

        self.init_member_list()
        self.init_account_combo_box()
        self.loader_movie.stop()
        self.loaderIconLabel.hide()

    def fetch_proposals_from_network(self):
        """
        Fetch proposals of technical committee from network

        :return:
        """
        try:
            self.application.technical_committee.network_update_proposals()
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_fetch_proposals_from_network_async_worker_finished(self):
        """
        Triggered when async worker is finished

        :return:
        """
        logging.debug("Technical committee widget proposals update")
        self.init_proposal_list()
        self.loader_movie.stop()
        self.loaderIconLabel.hide()

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        self.fetch_members_from_network_async_worker.start()
        self.fetch_proposals_from_network_async_worker.start()

    def on_vote_for_proposal_start(self):
        """
        Vote request send, display loader waiting for vote done

        :return:
        """
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.proposals_list_current_index = (
            self.proposalsListWidget.currentIndex().row()
        )

    def update_proposals(self):
        """
        After a vote, update proposals list with new votes

        :return:
        """
        self.fetch_proposals_from_network_async_worker.start()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    validator_node = Node(url="wss://gdev.cgeek.fr")
    if application_.nodes.get(validator_node.url) is None:
        application_.nodes.add(validator_node)
    application_.nodes.set_current_url(validator_node.url)
    application_.connections.node.connect(validator_node)
    main_window.setCentralWidget(
        TechnicalCommitteeWidget(application_, QMutex(), main_window)
    )

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, List, Optional, Union


class DBClientInterface(abc.ABC):
    """
    DBClientInterface class
    """

    @abc.abstractmethod
    def connect(self, uri: str) -> None:
        """
        Connect to database

        :param uri: Uri of file or server, or absolute path of file
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def disconnect(self) -> None:
        """
        Disconnect connection to database

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def execute(
        self, request: str, args: Optional[Union[List, tuple]] = None
    ) -> List[Any]:
        """
        Execute Select request returning an iterator on result set

        :param request: DB request
        :param args: Arguments of DB request
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def select(
        self, request: str, args: Optional[Union[List, tuple]] = None
    ) -> List[tuple]:
        """
        Execute Select request returning an iterator on result set

        :param request: DB request
        :param args: Arguments of DB request
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def insert(self, table: str, **kwargs):
        """
        Create a new entry in table, with field=value kwargs

        :param table: Table name
        :param kwargs: fields with their values
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def insert_many(self, table: str, entries: List[dict]):
        """
        Batch insert of entries in table, with each dict = {field: value}

        :param table: Table name
        :param entries: List of entries
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def select_one(
        self, sql: str, args: Optional[Union[List, tuple]] = None
    ) -> Optional[tuple]:
        """
        Execute SELECT sql query and return first result

        :param sql: SELECT query
        :param args: Query arguments
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, table: str, where: str, **kwargs):
        """
        Update rows of table selected by where from **kwargs

        :param table: Table name
        :param where: WHERE statement
        :param kwargs: field=values kwargs
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, table: str, **kwargs):
        """
        Delete a row from table where key=value (AND) from kwargs

        :param table: Table to delete from
        :param kwargs: Key/Value conditions (AND)
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self, table: str):
        """
        Clear table entries

        :param table: Name of the table
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import List, Optional

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QCloseEvent, QFont, QImage, QPixmap
from PyQt5.QtWidgets import QApplication, QDialog, QLabel, QVBoxLayout, QWidget
from scalecodec.utils.ss58 import is_valid_ss58_address

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import AccountEvent
from tikka.libs import crypto_type
from tikka.libs.keypair import Keypair
from tikka.slots.pyqt.entities.constants import ADDRESS_MONOSPACE_FONT_NAME
from tikka.slots.pyqt.resources.gui.windows.scan_qrcode_rc import Ui_ScanQRCodeDialog


class QrScannerWidget(QWidget):
    """QrScannerWidget class"""

    def __init__(self, camera_index: int = 0, parent: Optional[QWidget] = None):
        """
        Init QrScannerWidget instance

        :param camera_index: Index of camera from OpenCV detector
        :param parent: Parent QWidget
        """
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))
        self.image_label = QLabel(self)
        self.setup_ui()

        self.camera_index = camera_index
        self.detected_data = None
        self.cap = None
        self.timer = QTimer(self)

        self.setup_camera()

    def setup_ui(self):
        """
        Setup GUI

        :return:
        """
        self.image_label.setAlignment(Qt.AlignCenter)
        self.layout().addWidget(self.image_label)

    def setup_camera(self):
        """
        Setup camera with OpenCV

        :return:
        """
        import cv2

        self.cap = cv2.VideoCapture(self.camera_index)
        self.timer.timeout.connect(self.update_frame)
        self.timer.start(30)  # ~30 FPS

    def update_frame(self):
        """
        Update Qt frame from OpenCV camera images

        :return:
        """
        import cv2

        ret, frame = self.cap.read()
        if not ret:
            return

        # Detect QR code
        detector = cv2.QRCodeDetector()
        data, _, _ = detector.detectAndDecode(frame)

        if data and is_valid_ss58_address(data):
            self.detected_data = data
            self.timer.stop()
            self.parent().accept()  # Close parent QDialog

        # Display image
        rgb_image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        h, w, ch = rgb_image.shape
        bytes_per_line = ch * w
        qt_image = QImage(rgb_image.data, w, h, bytes_per_line, QImage.Format_RGB888)
        self.image_label.setPixmap(QPixmap.fromImage(qt_image))

    def closeEvent(self, event: QCloseEvent):
        """
        Oberload CloseEvent

        :param event: QCloseEvent instance
        :return:
        """
        if self.cap is not None:
            self.cap.release()
        super().closeEvent(event)


class ScanQRCodeOpenCVWindow(QDialog, Ui_ScanQRCodeDialog):
    """
    ScanQRCodeOpenCVWindow class
    """

    display_crypto_type = {
        AccountCryptoType.ED25519: "ED25519",
        AccountCryptoType.SR25519: "SR25519",
    }

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init scan qrcode OpenCV window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.scanner: Optional[QrScannerWidget] = None
        self.address: Optional[str] = None
        self.crypto_type = AccountCryptoType.ED25519

        # set monospace font to address field
        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)

        # buttons
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(False)

        # events
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)

        camera_index = self.get_available_camera_opencv_index()
        if camera_index is not None:
            self.start_qt_qr_scanner(camera_index)
        else:
            self.errorLabel.setText(self._("No camera available"))

    def start_qt_qr_scanner(self, camera_index: int):
        """
        Start OpenCV camera integrated in Qt

        :param camera_index: Index of camera
        :return:
        """
        self.scanner = QrScannerWidget(camera_index, self)
        self.layout().insertWidget(0, self.scanner)  # type: ignore

        # Hide useless elements during scan
        self.addressValueLabel.hide()
        self.keyTypeValueLabel.hide()
        # self.buttonBox.hide()

        if self.exec_() == QDialog.Accepted:
            self.address = self.scanner.detected_data
            if self.address is not None:
                self._show_result()
            else:
                self.errorLabel.setText(self._("QRCode Scanner canceled"))
        else:
            self.errorLabel.setText(self._("Scan canceled"))

    def _show_result(self):
        """
        Show scan result

        :return:
        """
        # self.scanner.hide()
        self.addressValueLabel.show()
        self.keyTypeValueLabel.show()
        # self.buttonBox.show()

        self.addressValueLabel.setText(self.address)
        self.errorLabel.setText("")
        self._detect_crypto_type(self.address)
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(True)

    def _detect_crypto_type(self, address: str):
        """
        :param address: Address scanned

        :return:
        """
        keypair = Keypair(address)
        try:
            result = crypto_type.is_valid_ed25519(keypair.public_key)
        except (AttributeError, AssertionError):
            result = False

        if result is True:
            self.crypto_type = AccountCryptoType.ED25519
            self.keyTypeValueLabel.setText(
                self.display_crypto_type[AccountCryptoType.ED25519]
            )
        else:
            try:
                result = crypto_type.is_valid_sr25519(keypair.public_key)
            except (AttributeError, AssertionError):
                result = False
            if result is True:
                self.crypto_type = AccountCryptoType.SR25519
                self.keyTypeValueLabel.setText(
                    self.display_crypto_type[AccountCryptoType.SR25519]
                )

    @staticmethod
    def get_available_camera_opencv_index() -> Optional[int]:
        """
        Return first available working camera index or None

        :return:
        """
        import cv2

        non_working_ports: List[int] = []
        dev_port = 0
        working_ports: List[int] = []
        available_ports: List[int] = []
        while len(non_working_ports) < 5:  # Stop after 5 non-working ports
            camera = cv2.VideoCapture(dev_port)
            if not camera.isOpened():
                non_working_ports.append(dev_port)
                logging.debug("Port %s is not working." % dev_port)
            else:
                is_reading, img = camera.read()
                width = camera.get(3)
                height = camera.get(4)
                if is_reading:
                    logging.debug(
                        "Camera %s is working and reads images (%s x %s)"
                        % (dev_port, width, height)
                    )
                    working_ports.append(dev_port)
                else:
                    logging.debug(
                        "Port %s for camera ( %s x %s) is present but does not reads."
                        % (dev_port, width, height)
                    )
                    available_ports.append(dev_port)
            dev_port += 1
            camera.release()

        return working_ports[0] if working_ports else None

    def on_accepted_button(self) -> None:
        """
        Triggered when user click on ok button

        :return:
        """
        if self.address is not None:
            account = self.application.accounts.get_by_address(self.address)
            if account is None:
                # create account instance
                account = Account(
                    self.address,
                    name=self._("Added from a QRCode"),
                    crypto_type=self.crypto_type,
                )
                # add instance in application
                self.application.accounts.add(account)
            else:
                # dispatch event
                event = AccountEvent(
                    AccountEvent.EVENT_TYPE_ADD,
                    account,
                )
                self.application.event_dispatcher.dispatch_event(event)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    window = ScanQRCodeOpenCVWindow(application_)
    if window.address is not None:
        # display window
        window.exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
import sqlite3
from pathlib import Path
from queue import Queue
from typing import Any, List, Optional, Union
from urllib.parse import urlparse

from yoyo import get_backend, read_migrations

from tikka.domains.entities.constants import DATABASE_MIGRATIONS_PATH
from tikka.interfaces.adapters.repository.db_client import DBClientInterface


class FileDBClient(DBClientInterface):
    """
    FileDBClient with connection pool to be threadsafe
    """

    CONNECTIONS_POOL_SIZE = 5

    def __init__(self):
        """
        Init a Sqlite3 database client adapter instance

        Use a connection pool system to be threadsafe and fast
        """
        super().__init__()

        self.uri: Optional[str] = None
        self.db_pool = Queue(self.CONNECTIONS_POOL_SIZE)

    def connect(self, uri: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DBClientInterface.connect.__doc__
        )
        self.uri = uri

        # update database
        self.migrate()

        # create pool of connections
        self.db_pool = Queue(self.CONNECTIONS_POOL_SIZE)
        for i in range(self.CONNECTIONS_POOL_SIZE):
            connection = sqlite3.connect(
                urlparse(self.uri).path, check_same_thread=False
            )
            connection.execute("PRAGMA foreign_keys = ON;")  # set check on foreign keys
            connection.commit()
            self.db_pool.put(connection)

    def disconnect(self):
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DBClientInterface.disconnect.__doc__
        )
        while not self.db_pool.not_empty:
            connection = self.db_pool.get()
            connection.close()
        logging.debug("Sqlite3 connection closed.")

    def execute(
        self, request: str, args: Optional[Union[List, tuple]] = None
    ) -> List[Any]:
        """
        Execute SQL request with arg

        :param request: SQL request
        :param args: Arguments of SQL request
        :return:
        """
        if args is None:
            args = tuple()

        # get a connection from pool
        connection = self.db_pool.get()

        cursor = connection.cursor()
        try:
            if args is None:
                cursor.execute(request)
            else:
                cursor.execute(request, args)
            data = cursor.fetchall()
        except Exception as exception:
            logging.debug(request)
            if args:
                logging.debug(args)
            raise exception
        finally:
            connection.commit()
            self.db_pool.put(connection)  # Return connection back to the pool

        return data

    def execute_many(
        self, request: str, args: Optional[List[tuple]] = None
    ) -> List[Any]:
        """
        Execute SQL request with arg

        :param request: SQL request
        :param args: Arguments of SQL request
        :return:
        """
        # get a connection from pool
        connection = self.db_pool.get()

        cursor = connection.cursor()
        try:
            if args is None:
                cursor.executemany(request)
            else:
                cursor.executemany(request, args)
            data = cursor.fetchall()
        finally:
            connection.commit()
            self.db_pool.put(connection)  # Return connection back to the pool

        return data

    def select(
        self, request: str, args: Optional[Union[List, tuple]] = None
    ) -> List[Any]:
        """
        Execute Select request returning an iterator on result set

        :param request: SQL request
        :param args: Arguments of SQL request
        :return:
        """
        return self.execute(request, args)

    def insert(self, table: str, **kwargs):
        """
        Create a new entry in table, with field=value kwargs

        :param table: Table name
        :param kwargs: fields with their values
        :return:
        """
        fields_string = ",".join(kwargs.keys())
        values_string = ",".join(["?" for _ in range(len(kwargs))])
        filtered_values = []
        for value in kwargs.values():
            # serialize dict to json string
            filtered_values.append(
                json.dumps(value) if isinstance(value, dict) else value
            )

        sql = (
            f"INSERT OR IGNORE INTO {table} ({fields_string}) VALUES ({values_string})"
        )
        self.execute(sql, filtered_values)

    def insert_many(self, table: str, entries: List[dict]):
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DBClientInterface.insert_many.__doc__
        )
        entry_values_list = []
        fields_string = ",".join(entries[0].keys())
        values_string = ",".join(["?" for _ in range(len(entries[0]))])
        for entry in entries:
            filtered_values = []
            for value in entry.values():
                # serialize dict to json string
                filtered_values.append(
                    json.dumps(value) if isinstance(value, dict) else value
                )
            entry_values_list.append(tuple(filtered_values))
        sql = (
            f"INSERT OR IGNORE INTO {table} ({fields_string}) VALUES ({values_string})"
        )
        self.execute_many(sql, entry_values_list)

    def select_one(
        self, sql: str, args: Optional[Union[List, tuple]] = None
    ) -> Optional[Any]:
        """
        Execute SELECT sql query and return first result

        :param sql: SELECT query
        :param args: Query arguments
        :return:
        """
        results = self.select(sql, args)
        if len(results) > 0:
            return results[0]

        return None

    def update(self, table: str, where: str, **kwargs):
        """
        Update rows of table selected by where from **kwargs

        :param table: Table name
        :param where: WHERE statement
        :param kwargs: field=values kwargs
        :return:
        """
        set_statement = ",".join([f"{field}=?" for field in kwargs])

        sql = f"UPDATE OR IGNORE {table} SET {set_statement} WHERE {where}"
        values = tuple(kwargs.values())

        self._update(sql, values)

    def _update(self, sql: str, values: tuple):
        """
        Send update request sql with values

        :param sql: SQL query
        :param values: Values to inject as sql params
        :return:
        """
        filtered_values = []
        for value in values:
            # serialize dict to json string
            filtered_values.append(
                json.dumps(value) if isinstance(value, dict) else value
            )
        self.execute(sql, filtered_values)

    def delete(self, table: str, **kwargs):
        """
        Delete a row from table where key=value (AND) from kwargs

        :param table: Table to delete from
        :param kwargs: Key/Value conditions (AND)
        :return:
        """
        conditions = " AND ".join([f"{key}=?" for key in kwargs])

        sql = f"DELETE FROM {table} WHERE {conditions}"
        self.execute(sql, tuple(kwargs.values()))

    def clear(self, table: str):
        """
        Clear table entries

        :param table: Name of the table
        :return:
        """
        # delete all entries
        self.execute(f"DELETE FROM {table}")

    def migrate(self):
        """
        Use Yoyo Python library to handle current database migrations

        :return:
        """
        migrations_path = str(
            Path(__file__).parent.parent.joinpath(DATABASE_MIGRATIONS_PATH).expanduser()
        )
        migrations = read_migrations(migrations_path)
        if self.uri is not None:

            backend = get_backend("sqlite:///" + urlparse(self.uri).path)
            with backend.lock():
                # Apply any outstanding migrations
                backend.apply_migrations(backend.to_apply(migrations))
                logging.debug(backend.applied_migrations_sql)


class NoConnectionError(Exception):
    pass



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from tikka.domains.accounts import Accounts
from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.constants import DERIVATION_SCAN_MAX_NUMBER
from tikka.interfaces.adapters.network.node.accounts import NodeAccountsInterface
from tikka.libs.keypair import Keypair


class Vaults:
    """
    Vaults domain class
    """

    def __init__(
        self,
        network: NodeAccountsInterface,
        accounts: Accounts,
        currencies: Currencies,
    ):
        """
        Init Vaults domain

        :param network: NetworkAccountsInterface instance
        :param accounts: Accounts domain instance
        :param currencies: Currencies domain instance
        """
        self.network = network
        self.accounts = accounts
        self.currencies = currencies

    def network_get_derived_accounts(
        self,
        root_address: str,
        mnemonic: str,
        language_code: str,
        crypto_type: AccountCryptoType,
    ) -> List[Account]:
        """
        Return list of derived accounts of root_address from network

        :param root_address: Root account address
        :param mnemonic: Mnemonic phrase
        :param language_code: Language code
        :param crypto_type: Crypto type

        :return:
        """
        derived_accounts: List[Account] = []
        addresses: List[str] = []
        for derivation_number in range(0, DERIVATION_SCAN_MAX_NUMBER + 1):
            derivation = f"//{derivation_number}"
            # # debug on Alice,...
            # if derivation == "//0":
            #     derivation = "//Alice"
            # if derivation == "//1":
            #     derivation = "//Bob"
            # if derivation == "//2":
            #     derivation = "//Charlie"
            # if derivation == "//3":
            #     derivation = "//Dave"

            suri = mnemonic + derivation
            keypair = Keypair.create_from_uri(
                suri,
                language_code=language_code,
                ss58_format=self.currencies.get_current().ss58_format,
                crypto_type=crypto_type,
            )
            derived_account = self.accounts.get_by_address(keypair.ss58_address)
            if derived_account is None:
                derived_account = self.accounts.get_instance(
                    keypair.ss58_address, f"{derivation}"
                )
            derived_account.root = root_address
            derived_account.path = derivation
            derived_accounts.append(derived_account)
            addresses.append(keypair.ss58_address)

        balances = self.accounts.network_get_balances(addresses)
        existing_accounts: List[Account] = []
        for derived_account in derived_accounts:
            if balances[derived_account.address] is not None:
                derived_account.balance = balances[derived_account.address].total  # type: ignore
                derived_account.balance_available = balances[derived_account.address].available  # type: ignore
                derived_account.balance_reserved = balances[derived_account.address].reserved  # type: ignore
                existing_accounts.append(derived_account)

        return existing_accounts



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QMutex, QPoint, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QMainWindow, QStyledItemDelegate, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import (
    ConnectionsEvent,
    CurrencyEvent,
    IndexersEvent,
    NodesEvent,
)
from tikka.domains.entities.indexer import Indexer
from tikka.domains.entities.node import Node
from tikka.slots.pyqt.entities.constants import (
    INDEXERS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    INDEXERS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    NODES_TABLE_SORT_ORDER_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.models.indexers import IndexersTableModel
from tikka.slots.pyqt.models.nodes import NodesTableModel
from tikka.slots.pyqt.resources.gui.widgets.servers_rc import Ui_ServersWidget
from tikka.slots.pyqt.widgets.indexer_menu import IndexerPopupMenu
from tikka.slots.pyqt.widgets.node_menu import NodePopupMenu
from tikka.slots.pyqt.windows.indexer_add import IndexerAddWindow
from tikka.slots.pyqt.windows.node_add import NodeAddWindow


class HighlightNodeDelegate(QStyledItemDelegate):
    """
    Class to highlight connected node row
    """

    def __init__(self, table_model, application):
        """
        Initialize HighlightNodeDelegate instance

        :param table_model: TableModel of
        :param application:
        """
        super().__init__()
        self.table_model = table_model
        self.application = application

    def paint(self, painter, option, index):
        # get current url/row node
        current_node_url = self.application.nodes.get_current_url()
        node = self.table_model.nodes[index.row()]
        # check connection...
        if node.url == current_node_url:
            color = (
                QColor("lightgreen")
                if self.application.connections.node.is_connected()
                else QColor("lightcoral")
            )
            painter.fillRect(option.rect, color)

        # paint rectangle
        super().paint(painter, option, index)


class HighlightIndexerDelegate(QStyledItemDelegate):
    """
    Class to highlight connected indexer row
    """

    def __init__(self, table_model, application):
        """
        Initialize HighlightIndexerDelegate instance

        :param table_model: TableModel of
        :param application:
        """
        super().__init__()
        self.table_model = table_model
        self.application = application

    def paint(self, painter, option, index):
        # get current url/row indexer
        current_indexer_url = self.application.indexers.get_current_url()
        indexer = self.table_model.indexers[index.row()]
        # check connection...
        if indexer.url == current_indexer_url:
            color = (
                QColor("lightgreen")
                if self.application.connections.indexer.is_connected()
                else QColor("lightcoral")
            )
            painter.fillRect(option.rect, color)

        # paint rectangle
        super().paint(painter, option, index)


class ServersWidget(QWidget, Ui_ServersWidget):
    """
    ServersWidget class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init ServersWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.mutex = mutex
        self._ = self.application.translator.gettext
        self.selected_node: Optional[Node] = None
        self.selected_indexer: Optional[Indexer] = None

        self.nodes_table_model = NodesTableModel(self.application)
        self.nodesTableView.setModel(self.nodes_table_model)
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()
        self.nodesTableView.customContextMenuRequested.connect(
            self.on_node_context_menu
        )
        self.nodesTableView.setItemDelegate(
            HighlightNodeDelegate(self.nodes_table_model, self.application)
        )
        pref_sort_column = self.application.repository.preferences.get(
            NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = 0
        else:
            sort_column = int(pref_sort_column)

        pref_sort_order = self.application.repository.preferences.get(
            NODES_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = int(Qt.SortOrder.AscendingOrder)
        else:
            sort_order = int(pref_sort_order)
        self.nodesTableView.sortByColumn(sort_column, sort_order)

        self.indexers_table_model = IndexersTableModel(self.application)
        self.indexersTableView.setModel(self.indexers_table_model)
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()
        self.indexersTableView.customContextMenuRequested.connect(
            self.on_indexer_context_menu
        )
        self.indexersTableView.setItemDelegate(
            HighlightIndexerDelegate(self.indexers_table_model, self.application)
        )
        pref_sort_column = self.application.repository.preferences.get(
            INDEXERS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = 0
        else:
            sort_column = int(pref_sort_column)

        pref_sort_order = self.application.repository.preferences.get(
            INDEXERS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = int(Qt.SortOrder.AscendingOrder)
        else:
            sort_order = int(pref_sort_order)
        self.indexersTableView.sortByColumn(sort_column, sort_order)

        # events
        self.addNodeButton.clicked.connect(self._on_add_node_button_clicked)
        self.addIndexerButton.clicked.connect(self._on_add_indexer_button_clicked)
        self.nodesTableView.doubleClicked.connect(self.on_node_table_double_click)
        self.indexersTableView.doubleClicked.connect(self.on_indexer_table_double_click)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            NodesEvent.EVENT_TYPE_LIST_CHANGED, self._on_node_list_changed_event
        )
        self.application.event_dispatcher.add_event_listener(
            IndexersEvent.EVENT_TYPE_LIST_CHANGED, self._on_indexer_list_changed_event
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_NODE_CONNECTED,
            lambda x: self.nodesTableView.viewport().update(),
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_NODE_DISCONNECTED,
            lambda x: self.nodesTableView.viewport().update(),
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED,
            lambda x: self.indexersTableView.viewport().update(),
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_INDEXER_DISCONNECTED,
            lambda x: self.indexersTableView.viewport().update(),
        )

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.toggle_node_connection_async_qworker = AsyncQWorker(
            self.toggle_node_connection,
            self.mutex,
        )
        self.toggle_node_connection_async_qworker.finished.connect(
            self._on_finished_toggle_node_connection
        )
        self.toggle_indexer_connection_async_qworker = AsyncQWorker(
            self.toggle_indexer_connection,
            self.mutex,
        )
        self.toggle_indexer_connection_async_qworker.finished.connect(
            self._on_finished_toggle_indexer_connection
        )

    def _on_add_node_button_clicked(self):
        """
        Trigger when user click on add node button

        :return:
        """
        NodeAddWindow(self.application, self.mutex, self).exec_()

    def _on_add_indexer_button_clicked(self):
        """
        Trigger when user click on add indexer button

        :return:
        """
        IndexerAddWindow(self.application, self).exec_()

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        self.indexersTableView.model().init_data()

        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()

    def _on_node_list_changed_event(self, _):
        """
        When the node list has changed

        :param _: NodesEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()

    def _on_indexer_list_changed_event(self, _):
        """
        When the indexer list has changed

        :param _: IndexersEvent instance
        :return:
        """
        # update model
        self.indexersTableView.model().init_data()
        # resize view
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()

    def on_node_context_menu(self, position: QPoint):
        """
        When right button on node table view

        :param position: QPoint instance
        :return:
        """
        index = self.nodesTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            node = self.nodes_table_model.nodes[row]
            # display popup menu at click position
            NodePopupMenu(self.application, node).exec_(
                self.nodesTableView.mapToGlobal(position)
            )

    def on_indexer_context_menu(self, position: QPoint):
        """
        When right button on indexer table view

        :param position: QPoint instance
        :return:
        """
        index = self.indexersTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            indexer = self.indexers_table_model.indexers[row]
            # display popup menu at click position
            IndexerPopupMenu(self.application, indexer).exec_(
                self.indexersTableView.mapToGlobal(position)
            )

    def on_node_table_double_click(self, index):
        """
        Triggered when user double-click on a row of node table view
        """
        row = index.row()
        # check index...
        if 0 <= row < len(self.nodes_table_model.nodes):
            self.selected_node = self.nodes_table_model.nodes[row]
            self.toggle_node_connection_async_qworker.start()

    def toggle_node_connection(self):
        """
        Toggle node connection

        :return:
        """
        if self.selected_node is None:
            return
        # current url connected
        current_node_url = self.application.nodes.get_current_url()

        # check connection...
        if (
            self.selected_node.url == current_node_url
            and self.application.connections.node.is_connected()
        ):
            self.application.connections.node.disconnect()
        else:
            if self.selected_node.url != current_node_url:
                self.application.nodes.set_current_url(self.selected_node.url)

            self.application.connections.node.connect(self.selected_node)
            if self.application.connections.node.is_connected():
                node_currency = self.application.currencies.network.get_instance()
                if (
                    self.application.currencies.get_current().genesis_hash is not None
                    and node_currency.genesis_hash
                    != self.application.currencies.get_current().genesis_hash
                ):
                    self.application.connections.node.disconnect()
                    logging.error("Node currency is different! Force Disconnect!")

    def _on_finished_toggle_node_connection(self):
        """
        Triggered when toggle_node_connection is finished

        :return:
        """
        if self.application.connections.node.is_connected():
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_NODE_CONNECTED)
            )
        else:
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_NODE_DISCONNECTED)
            )

    def on_indexer_table_double_click(self, index):
        """
        Triggered when user double-click on a row of indexer table view
        """
        row = index.row()
        # check index...
        if 0 <= row < len(self.indexers_table_model.indexers):
            self.selected_indexer = self.indexers_table_model.indexers[row]
            self.toggle_indexer_connection_async_qworker.start()

    def toggle_indexer_connection(self):
        """
        Toggle indexer connection

        :return:
        """
        if self.selected_indexer is None:
            return
        # current url connected
        current_indexer_url = self.application.indexers.get_current_url()

        # check connection...
        if (
            self.selected_indexer.url == current_indexer_url
            and self.application.connections.indexer.is_connected()
        ):
            self.application.connections.indexer.disconnect()
        else:
            if self.selected_indexer.url != current_indexer_url:
                self.application.indexers.set_current_url(self.selected_indexer.url)

            self.application.connections.indexer.connect(self.selected_indexer)
            if self.application.connections.indexer.is_connected():
                indexer_genesis_hash = (
                    self.application.indexers.network_get_genesis_hash()
                )
                if (
                    self.application.currencies.get_current().genesis_hash is not None
                    and indexer_genesis_hash
                    != self.application.currencies.get_current().genesis_hash
                ):
                    self.application.connections.indexer.disconnect()
                    logging.error("Indexer currency is different! Force Disconnect!")

    def _on_finished_toggle_indexer_connection(self):
        """
        Triggered when toggle_indexer_connection is finished

        :return:
        """
        if self.application.connections.indexer.is_connected():
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED)
            )
        else:
            self.application.event_dispatcher.dispatch_event(
                ConnectionsEvent(ConnectionsEvent.EVENT_TYPE_INDEXER_DISCONNECTED)
            )


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(ServersWidget(application_, QMutex(), main_window))

    sys.exit(qapp.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/widgets/connection.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ConnectionWidget(object):
    def setupUi(self, ConnectionWidget):
        ConnectionWidget.setObjectName("ConnectionWidget")
        ConnectionWidget.resize(724, 424)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(ConnectionWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.groupBox = QtWidgets.QGroupBox(ConnectionWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.unsafeAPIExposedValueLabel = QtWidgets.QLabel(self.groupBox)
        self.unsafeAPIExposedValueLabel.setText("")
        self.unsafeAPIExposedValueLabel.setObjectName("unsafeAPIExposedValueLabel")
        self.gridLayout.addWidget(self.unsafeAPIExposedValueLabel, 7, 1, 1, 1)
        self.unsafeAPIExposedLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.unsafeAPIExposedLabel.setFont(font)
        self.unsafeAPIExposedLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.unsafeAPIExposedLabel.setObjectName("unsafeAPIExposedLabel")
        self.gridLayout.addWidget(self.unsafeAPIExposedLabel, 7, 0, 1, 1)
        self.nodeBlockValueLabel = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nodeBlockValueLabel.sizePolicy().hasHeightForWidth())
        self.nodeBlockValueLabel.setSizePolicy(sizePolicy)
        self.nodeBlockValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);")
        self.nodeBlockValueLabel.setObjectName("nodeBlockValueLabel")
        self.gridLayout.addWidget(self.nodeBlockValueLabel, 5, 1, 1, 1)
        self.nodeBlockLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nodeBlockLabel.setFont(font)
        self.nodeBlockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nodeBlockLabel.setObjectName("nodeBlockLabel")
        self.gridLayout.addWidget(self.nodeBlockLabel, 5, 0, 1, 1)
        self.versionLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.versionLabel.setFont(font)
        self.versionLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.versionLabel.setObjectName("versionLabel")
        self.gridLayout.addWidget(self.versionLabel, 3, 0, 1, 1)
        self.epochValueLabel = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.epochValueLabel.sizePolicy().hasHeightForWidth())
        self.epochValueLabel.setSizePolicy(sizePolicy)
        self.epochValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);")
        self.epochValueLabel.setObjectName("epochValueLabel")
        self.gridLayout.addWidget(self.epochValueLabel, 6, 1, 1, 1)
        self.nodeConnectButton = QtWidgets.QPushButton(self.groupBox)
        self.nodeConnectButton.setObjectName("nodeConnectButton")
        self.gridLayout.addWidget(self.nodeConnectButton, 1, 3, 1, 1)
        self.versionValueLabel = QtWidgets.QLabel(self.groupBox)
        self.versionValueLabel.setText("")
        self.versionValueLabel.setObjectName("versionValueLabel")
        self.gridLayout.addWidget(self.versionValueLabel, 3, 1, 1, 1)
        self.softwareLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.softwareLabel.setFont(font)
        self.softwareLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.softwareLabel.setObjectName("softwareLabel")
        self.gridLayout.addWidget(self.softwareLabel, 2, 0, 1, 1)
        self.softwareValueLabel = QtWidgets.QLabel(self.groupBox)
        self.softwareValueLabel.setText("")
        self.softwareValueLabel.setObjectName("softwareValueLabel")
        self.gridLayout.addWidget(self.softwareValueLabel, 2, 1, 1, 1)
        self.peerIDLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.peerIDLabel.setFont(font)
        self.peerIDLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.peerIDLabel.setObjectName("peerIDLabel")
        self.gridLayout.addWidget(self.peerIDLabel, 4, 0, 1, 1)
        self.nodeUrlLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nodeUrlLabel.setFont(font)
        self.nodeUrlLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nodeUrlLabel.setObjectName("nodeUrlLabel")
        self.gridLayout.addWidget(self.nodeUrlLabel, 1, 0, 1, 1)
        self.peerIDValueLabel = QtWidgets.QLabel(self.groupBox)
        self.peerIDValueLabel.setText("")
        self.peerIDValueLabel.setObjectName("peerIDValueLabel")
        self.gridLayout.addWidget(self.peerIDValueLabel, 4, 1, 1, 1)
        self.epochLabel = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.epochLabel.setFont(font)
        self.epochLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.epochLabel.setObjectName("epochLabel")
        self.gridLayout.addWidget(self.epochLabel, 6, 0, 1, 1)
        self.nodeUrlsComboBox = QtWidgets.QComboBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nodeUrlsComboBox.sizePolicy().hasHeightForWidth())
        self.nodeUrlsComboBox.setSizePolicy(sizePolicy)
        self.nodeUrlsComboBox.setToolTip("")
        self.nodeUrlsComboBox.setEditable(False)
        self.nodeUrlsComboBox.setMinimumContentsLength(0)
        self.nodeUrlsComboBox.setObjectName("nodeUrlsComboBox")
        self.gridLayout.addWidget(self.nodeUrlsComboBox, 1, 1, 1, 1)
        self.nodeConnectionStatusLabel = QtWidgets.QLabel(self.groupBox)
        self.nodeConnectionStatusLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.nodeConnectionStatusLabel.setText("")
        self.nodeConnectionStatusLabel.setPixmap(QtGui.QPixmap(":/icons/disconnected"))
        self.nodeConnectionStatusLabel.setScaledContents(True)
        self.nodeConnectionStatusLabel.setObjectName("nodeConnectionStatusLabel")
        self.gridLayout.addWidget(self.nodeConnectionStatusLabel, 1, 2, 1, 1)
        self.refreshNodeButton = QtWidgets.QPushButton(self.groupBox)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshNodeButton.setIcon(icon)
        self.refreshNodeButton.setObjectName("refreshNodeButton")
        self.gridLayout.addWidget(self.refreshNodeButton, 1, 4, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.verticalLayout_2.addWidget(self.groupBox, 0, QtCore.Qt.AlignTop)
        self.groupBox_2 = QtWidgets.QGroupBox(ConnectionWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.indexerUrlsComboBox = QtWidgets.QComboBox(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.indexerUrlsComboBox.sizePolicy().hasHeightForWidth())
        self.indexerUrlsComboBox.setSizePolicy(sizePolicy)
        self.indexerUrlsComboBox.setObjectName("indexerUrlsComboBox")
        self.gridLayout_2.addWidget(self.indexerUrlsComboBox, 0, 1, 1, 1)
        self.indexerUrlLabel = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.indexerUrlLabel.sizePolicy().hasHeightForWidth())
        self.indexerUrlLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.indexerUrlLabel.setFont(font)
        self.indexerUrlLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.indexerUrlLabel.setObjectName("indexerUrlLabel")
        self.gridLayout_2.addWidget(self.indexerUrlLabel, 0, 0, 1, 1)
        self.indexerConnectButton = QtWidgets.QPushButton(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.indexerConnectButton.sizePolicy().hasHeightForWidth())
        self.indexerConnectButton.setSizePolicy(sizePolicy)
        self.indexerConnectButton.setObjectName("indexerConnectButton")
        self.gridLayout_2.addWidget(self.indexerConnectButton, 0, 3, 1, 1)
        self.indexerBlockLabel = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.indexerBlockLabel.setFont(font)
        self.indexerBlockLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.indexerBlockLabel.setObjectName("indexerBlockLabel")
        self.gridLayout_2.addWidget(self.indexerBlockLabel, 1, 0, 1, 1)
        self.indexerBlockValueLabel = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.indexerBlockValueLabel.sizePolicy().hasHeightForWidth())
        self.indexerBlockValueLabel.setSizePolicy(sizePolicy)
        self.indexerBlockValueLabel.setStyleSheet("color: white;\n"
"border : 3px solid rgb(34, 156, 244);\n"
"border-radius: 10%;\n"
"background: rgb(34, 156, 244);")
        self.indexerBlockValueLabel.setObjectName("indexerBlockValueLabel")
        self.gridLayout_2.addWidget(self.indexerBlockValueLabel, 1, 1, 1, 1)
        self.indexerConnectionStatusLabel = QtWidgets.QLabel(self.groupBox_2)
        self.indexerConnectionStatusLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.indexerConnectionStatusLabel.setText("")
        self.indexerConnectionStatusLabel.setPixmap(QtGui.QPixmap(":/icons/disconnected"))
        self.indexerConnectionStatusLabel.setScaledContents(True)
        self.indexerConnectionStatusLabel.setObjectName("indexerConnectionStatusLabel")
        self.gridLayout_2.addWidget(self.indexerConnectionStatusLabel, 0, 2, 1, 1)
        self.refreshIndexerButton = QtWidgets.QPushButton(self.groupBox_2)
        self.refreshIndexerButton.setIcon(icon)
        self.refreshIndexerButton.setObjectName("refreshIndexerButton")
        self.gridLayout_2.addWidget(self.refreshIndexerButton, 0, 4, 1, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.retranslateUi(ConnectionWidget)
        QtCore.QMetaObject.connectSlotsByName(ConnectionWidget)

    def retranslateUi(self, ConnectionWidget):
        #  _translate = QtCore.QCoreApplication.translate
        ConnectionWidget.setWindowTitle(_("Connection"))
        self.groupBox.setTitle(_("Node"))
        self.unsafeAPIExposedLabel.setText(_("Unsafe API exposed"))
        self.nodeBlockValueLabel.setText(_("?"))
        self.nodeBlockLabel.setText(_("Block"))
        self.versionLabel.setText(_("Version"))
        self.epochValueLabel.setText(_("?"))
        self.nodeConnectButton.setToolTip(_("Connection to node"))
        self.nodeConnectButton.setText(_("Connect"))
        self.softwareLabel.setText(_("Software"))
        self.peerIDLabel.setText(_("Peer ID"))
        self.nodeUrlLabel.setText(_("Url"))
        self.epochLabel.setText(_("Epoch"))
        self.refreshNodeButton.setToolTip(_("Refresh node infos from network"))
        self.refreshNodeButton.setText(_("Refresh"))
        self.groupBox_2.setTitle(_("Indexer"))
        self.indexerUrlLabel.setText(_("Url"))
        self.indexerConnectButton.setToolTip(_("Connection to the indexer"))
        self.indexerConnectButton.setText(_("Connect"))
        self.indexerBlockLabel.setText(_("Block"))
        self.indexerBlockValueLabel.setText(_("?"))
        self.refreshIndexerButton.setToolTip(_("Refresh indexer infos from network"))
        self.refreshIndexerButton.setText(_("Refresh"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import socket
from pathlib import Path
from typing import Optional

from tikka.adapters.network.node.substrate_client import SubstrateClient
from tikka.domains.entities.constants import DATA_PATH
from tikka.interfaces.adapters.network.connection import ConnectionInterface
from tikka.interfaces.entities.Server import ServerInterface

# websocket timeout
RPC_CONNECTION_TIMEOUT = 30


class NodeConnection(ConnectionInterface):
    """
    NodeConnection class
    """

    def __init__(self) -> None:
        """
        Init NodeConnection instance

        RPC client is available in self.client after connect()
        """
        self.client: Optional[SubstrateClient] = None

    def connect(self, server: ServerInterface) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.connect.__doc__
        )
        self.url = server.url
        logging.debug(
            "CONNECTING TO NODE %s......................................", server.url
        )
        try:
            self.client = SubstrateClient(
                url=server.url,
                websocket_options={
                    "timeout": RPC_CONNECTION_TIMEOUT,
                    "sockopt": ((socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),),
                },
                metadata_dir=str(Path(DATA_PATH).expanduser()),
            )
        except Exception as exception:
            self.client = None
            logging.exception(exception)
            logging.debug(
                "CONNECTION TO NODE %s......................................FAILED !",
                server.url,
            )

        if self.client is not None:
            logging.debug(
                "CONNECTION TO NODE %s......................................SUCCESS !",
                server.url,
            )

    def disconnect(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.disconnect.__doc__
        )
        if self.client is not None:
            self.client.close()
            self.client = None

    def is_connected(self) -> bool:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            ConnectionInterface.is_connected.__doc__
        )
        return self.client is not None



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

import qrcode
from PyQt5.QtCore import QEvent, QMutex, QPoint, pyqtSignal
from PyQt5.QtGui import QFont, QPixmap, QWheelEvent
from PyQt5.QtWidgets import QApplication, QListView, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import (
    AccountEvent,
    ConnectionsEvent,
    TransferEvent,
    UnitEvent,
)
from tikka.domains.entities.identity import IdentityStatus
from tikka.domains.entities.smith import SmithStatus
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_IDENTITY_MEMBER,
    ICON_IDENTITY_MEMBER_NOT_VALIDATED,
    ICON_IDENTITY_NOT_MEMBER,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.qrcode_image import QRCodeImage
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.models.transfers import TransferItemDelegate, TransfersListModel
from tikka.slots.pyqt.resources.gui.widgets.account_rc import Ui_AccountWidget
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu
from tikka.slots.pyqt.widgets.transfer_history_menu import TransferHistoryPopupMenu
from tikka.slots.pyqt.windows.transfer import TransferWindow


class TransfersListView(QListView):
    """
    TransfersListView class
    """

    scroll_offset_changed = pyqtSignal(int)

    def __init__(
        self,
        application: Application,
        address: str,
        page_size: int,
        parent: Optional[QWidget] = None,
    ):
        """
        Init TransfersListView instance

        :param application: Application instance
        :param address: Account address
        :param page_size: Nb of transfers per page
        :param parent: QWidget instance, default None
        """
        super().__init__(parent)

        self.application = application
        self.address = address
        self.scroll_offset = 0
        self.page_size = page_size

    def wheelEvent(self, event: QWheelEvent):
        """
        Override Wheel event

        :param event: QEvent instance
        :return:
        """
        delta = event.angleDelta().y()
        if not self.verticalScrollBar().isVisible():
            if delta < 0:  # 🔽 Scroll vers le bas (charger plus)
                if (
                    self.scroll_offset + self.page_size
                    < self.application.transfers.count(self.address)
                ):
                    self.scroll_offset += self.page_size
                    self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                    self.scroll_offset_changed.emit(self.scroll_offset)

            elif delta > 0:  # 🔼 Scroll vers le haut (charger page précédente si dispo)
                self.scroll_offset -= self.page_size
                if self.scroll_offset < 0:
                    self.scroll_offset = 0
                self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                self.scroll_offset_changed.emit(self.scroll_offset)
        else:
            # wheel backward and scroll bar maximum down...
            if (
                delta < 0
                and self.verticalScrollBar().value()
                == self.verticalScrollBar().maximum()
            ):

                if (
                    self.scroll_offset + self.page_size
                    < self.application.transfers.count(self.address)
                ):
                    self.scroll_offset += self.page_size
                    self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                    self.scroll_offset_changed.emit(self.scroll_offset)

            # wheel forward and scroll bar maximum top...
            elif (
                delta > 0
                and self.verticalScrollBar().value()
                == self.verticalScrollBar().minimum()
            ):
                self.scroll_offset -= self.page_size
                if self.scroll_offset < 0:
                    self.scroll_offset = 0
                self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                self.scroll_offset_changed.emit(self.scroll_offset)

        super().wheelEvent(event)


class AccountWidget(QWidget, Ui_AccountWidget):
    """
    AccountWidget class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountWidget instance

        :param application: Application instance
        :param account: Account instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.account = account
        self.mutex = mutex
        self.history_limit = 100
        self.history_page_size = 100
        self.history_scroll_offset = 0

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(self.monospace_font)
        self.addressValueLabel.setText(self.account.address)
        if self.account.legacy_v1 is True:
            self.v1AddressValueLabel.setText(
                self.account.get_v1_address(
                    self.application.currencies.get_current().ss58_format
                )
            )

        # creating a pix map of qr code
        qr_code_pixmap = qrcode.make(
            self.account.address, image_factory=QRCodeImage
        ).pixmap()

        # set qrcode to the label
        self.QRCodeAddressLabel.setPixmap(qr_code_pixmap)

        self.transfers_model = TransfersListModel(
            application, self.account.address, self.locale()
        )

        # self.transfersListView.hide()
        # self.transfersListView = TransfersListView(
        #     application, self.account.address, self.history_page_size, self.groupBox
        # )
        #
        # self.transfersListView.setMouseTracking(False)
        # self.transfersListView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        # self.transfersListView.setAlternatingRowColors(True)
        # self.transfersListView.setSelectionMode(QAbstractItemView.NoSelection)
        # self.transfersListView.setObjectName("transfersListView")
        # self.verticalLayout_4.addWidget(self.transfersListView)

        self.transfersListView.setModel(self.transfers_model)
        self.transfersListView.setItemDelegate(TransferItemDelegate())
        self.transfersListView.setResizeMode(QListView.Adjust)

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_from_network_async_qworker = AsyncQWorker(
            self.fetch_account_from_network, self.mutex
        )
        self.fetch_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_from_network
        )
        self.fetch_transfers_from_network_async_qworker = AsyncQWorker(
            self.fetch_transfers_from_network, self.mutex
        )
        self.fetch_transfers_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_transfers_from_network
        )

        # events
        self.refreshButton.clicked.connect(self._on_refresh_button_clicked)
        self.transferToButton.clicked.connect(self.transfer)
        self.customContextMenuRequested.connect(self.on_context_menu)
        # self.transfersListView.scroll_offset_changed.connect(self._update_ui)
        self.transfersListView.customContextMenuRequested.connect(
            self.on_transfer_history_context_menu
        )
        # application events
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, lambda e: self._update_ui()
        )
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, lambda e: self._on_unit_changed_event()
        )
        self.application.event_dispatcher.add_event_listener(
            TransferEvent.EVENT_TYPE_SENT, lambda e: self._on_transfer_sent()
        )
        self.application.event_dispatcher.add_event_listener(
            ConnectionsEvent.EVENT_TYPE_INDEXER_CONNECTED,
            lambda e: self._on_indexer_connected(),
        )

        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )
        self._update_ui()

    def transfer(self):
        """
        When user click on transfer to button

        :return:
        """
        TransferWindow(
            self.application, self.mutex, None, self.account, parent=self.parent()
        ).exec_()

    def _on_refresh_button_clicked(self, _: QEvent):
        """
        Triggered when user click on Refresh button

        :param _: QEvent instance
        :return:
        """
        self.errorLabel.setText("")
        self.fetch_from_network_async_qworker.start()
        self.fetch_transfers_from_network_async_qworker.start()

    def _on_indexer_connected(self):
        """ "
        Triggered when the indexer is connected
        """
        self.fetch_transfers_from_network_async_qworker.start()

    def _on_transfer_sent(self):
        """
        Triggered after a successful transfer event

        :return:
        """
        # update account balance from network
        self.fetch_from_network_async_qworker.start()
        self.fetch_transfers_from_network_async_qworker.start()

    def _on_unit_changed_event(self):
        """
        Triggered when user change money units

        :return:
        """
        self._update_ui()
        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )

    def fetch_account_from_network(self):
        """
        Get last account data from the network

        :return:
        """
        if not self.application.connections.node.is_connected():
            return

        self.refreshButton.setEnabled(False)

        try:
            self.application.accounts.network_update_account(self.account)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            try:
                identity = self.application.identities.network_update_identity(
                    self.account.address
                )
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                if identity is not None:
                    try:
                        self.application.smiths.network_update_smith(identity.index)
                    except Exception as exception:
                        self.errorLabel.setText(self._(str(exception)))
                        logging.exception(exception)

    def _on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_account_from_network is finished

        :return:
        """
        self.account = self.application.accounts.get_by_address(self.account.address)
        self._update_ui()
        self.refreshButton.setEnabled(True)
        logging.debug("Account widget update")

    def fetch_transfers_from_network(self):
        """
        Get transfers data from the network

        :return:
        """
        if not self.application.connections.indexer.is_connected():
            return

        self.refreshButton.setEnabled(False)

        try:
            self.application.transfers.network_fetch_history_for_account(
                self.account, self.history_limit
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

        try:
            self.account.total_transfers_count = (
                self.application.transfers.network_fetch_total_count_for_address(
                    self.account.address
                )
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_fetch_transfers_from_network(self):
        """
        Triggered when async request fetch_transfers_from_network is finished

        :return:
        """
        self.application.accounts.update(self.account)

        self.account = self.application.accounts.get_by_address(self.account.address)
        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )
        self._update_ui()
        self.refreshButton.setEnabled(True)
        logging.debug(
            "Finished fetch account transfers %s %s",
            self.account.name,
            self.account.address,
        )

    def _update_ui(self):
        """
        Update UI from self.account

        :return:
        """
        display_identity_status = {
            IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
            IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
            IdentityStatus.MEMBER.value: self._("Member"),
            IdentityStatus.NOT_MEMBER.value: self._("Not member"),
            IdentityStatus.REVOKED.value: self._("Revoked"),
        }

        display_identity_icon = {
            IdentityStatus.UNCONFIRMED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.UNVALIDATED.value: QPixmap(
                ICON_IDENTITY_MEMBER_NOT_VALIDATED
            ),
            IdentityStatus.MEMBER.value: QPixmap(ICON_IDENTITY_MEMBER),
            IdentityStatus.NOT_MEMBER.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
            IdentityStatus.REVOKED.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
        }

        display_smith_status = {
            SmithStatus.INVITED.value: self._("Invited"),
            SmithStatus.PENDING.value: self._("Pending"),
            SmithStatus.SMITH.value: self._("Smith"),
            SmithStatus.EXCLUDED.value: self._("Excluded"),
        }

        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            amount = self.application.amounts.get_amount(unit_preference)
        else:
            amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        if self.account.balance is None:
            self.balanceValueLabel.setText("?")
            self.balanceReservedValueLabel.setText("")
        else:
            self.balanceValueLabel.setText(
                self.locale().toCurrencyString(
                    amount.value(self.account.balance), amount.symbol()
                )
            )
            display_reserved_balance = self.locale().toCurrencyString(
                amount.value(self.account.balance_reserved), amount.symbol()
            )
            self.balanceReservedValueLabel.setText(f"[-{display_reserved_balance}]")

        if self.application.wallets.exists(self.account.address):
            if self.application.wallets.is_unlocked(self.account.address):
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED))
            else:
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_LOCKED))
        else:
            self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_NO_WALLET))

        identity = self.application.identities.get_by_address(self.account.address)
        if identity is not None:
            self.identityStatusIconLabel.setPixmap(
                display_identity_icon[identity.status.value]
            )
            identity_name = identity.name or ""
            self.identityNameValueLabel.setText(f"{identity_name}#{identity.index}")
            self.identityValueLabel.setText(
                display_identity_status[identity.status.value]
            )
            smith = self.application.smiths.get(identity.index)
            if smith is not None:
                self.smithValuelabel.setText(display_smith_status[smith.status.value])
            else:
                self.smithValuelabel.setText(self._("No"))
            if self.account.name is None:
                self.accountNameValueLabel.setText("")
            else:
                self.accountNameValueLabel.setText(f" - {self.account.name}")
        else:
            self.identityStatusIconLabel.setPixmap(QPixmap(ICON_IDENTITY_NOT_MEMBER))
            self.identityNameValueLabel.setText("")
            self.identityValueLabel.setText(self._("No"))
            self.smithValuelabel.setText(self._("No"))
            if self.account.name is None:
                self.accountNameValueLabel.setText("")
            else:
                self.accountNameValueLabel.setText(self.account.name)

        if self.account.root is None and self.account.path is None:
            self.derivationValueLabel.setText(self._("Root"))
        else:
            root_account = self.application.accounts.get_by_address(self.account.root)
            if root_account is not None and root_account.name is not None:
                self.derivationValueLabel.setFont(QFont())
                self.derivationValueLabel.setText(root_account.name + self.account.path)
            else:
                self.derivationValueLabel.setFont(self.monospace_font)
                self.derivationValueLabel.setText(
                    DisplayAddress(self.account.root).shorten + self.account.path
                )

        # transfers
        self.transfersCountValueLabel.setText(
            f"1...{self.history_limit}/{self.account.total_transfers_count}"
        )
        # current_page = int(
        #     (self.transfersListView.scroll_offset / self.history_page_size) + 1
        # )
        # self.transfersCountValueLabel.setText(
        #     str(
        #         f"Page {current_page}/{self.history_page_size} |"
        #         f"1...{self.history_limit}/{self.account.total_transfers_count}"
        #     )
        # )

    def on_context_menu(self, position: QPoint):
        """
        When right button on account tab

        :return:
        """
        # display popup menu at click position
        menu = AccountPopupMenu(self.application, self.account, self.mutex, self)
        menu.exec_(self.mapToGlobal(position))

    def on_transfer_history_context_menu(self, position: QPoint):
        """
        When right button on transfer history listview

        :return:
        """
        # get selected transfer
        transfer = self.transfersListView.currentIndex().internalPointer()
        if transfer is not None:
            # display popup menu at click position
            menu = TransferHistoryPopupMenu(
                self.application, self.account, transfer, self.mutex, self
            )
            menu.exec_(self.transfersListView.mapToGlobal(position))


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = Account("5GrwvaEF5zXb26Fz9rcQpDWS57CtERHpNehXCPcNoHGKutQY")
    account_.balance = 1000
    account_.balance_available = 900
    account_.balance_reserved = 100

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(
        AccountWidget(application_, account_, QMutex(), main_window)
    )

    sys.exit(qapp.exec_())
